from typing import Any, Dict
import os
import simple_file_handler as sfh
from dotenv import find_dotenv, dotenv_values
import dcrypt

from .exceptions import CryptKeysNotFound
from .misc import (
    _print, env_variables, find_and_load_env_var, 
    generate_django_secret_key, validate_secret_key
)


def _get_env_hdl():
    """Returns an `FileHandler` object for an .env file"""
    return sfh.FileHandler(find_dotenv(raise_error_if_not_found=True), allow_any=True) 



class DjangoJSONSecretManager:
    cryptkeys_filename = 'cryptkeys.json'
    """The default name in which the manager's cryptkeys will be stored"""

    def __init__(self, path_to_secret_file: str, *, quiet: bool = False) -> None:
        """
        Creates a new DjangoJSONSecretManager object.

        :param path_to_secret_file: path the file where secrets are stored or secrets will be stored.
        :param quiet: If True, do not write anything to stdout (to avoid cluttering the console)
        """
        try:
            path = os.path.normpath(path_to_secret_file)
        except Exception:
            raise ValueError('`path_to_secret_file` must be a valid path')
        if not path.endswith('.json'):
            raise ValueError(
                'Secret file must be a json file'
            )
        self.secrets_file_path = path
        self.stdout = lambda msg: _print(msg, quiet=quiet)
        
        dj_keyname = os.getenv('DJSM_SECRET_KEY_NAME')
        if dj_keyname:
            if not isinstance(dj_keyname, str):
                raise TypeError('DJSM_SECRET_KEY_NAME must be a string')
            if not dj_keyname.isidentifier():
                raise ValueError('DJSM_SECRET_KEY_NAME must be a valid identifier')
            self.secret_key_name = dj_keyname
        else:
            self.secret_key_name = 'django_secret_key'
        return None

    @property
    def cryptkeys_filepath(self) -> str:
        """Returns the path to the crypt keys file"""
        return f"{os.path.dirname(self.secrets_file_path)}\\{self.cryptkeys_filename.split('.')[0]}.json"
    

    def __setattr__(self, __name: str, __value: Any) -> None:
        if "path" in __name:
            if not os.path.isabs(__value):
                __value = os.path.abspath(__value)
        return super().__setattr__(__name, __value)


    def _encrypt(self, secret: Dict) -> Dict:
        """
        Encrypts a secret using the encryption key.

        :param secret: secret to be encrypted
        :return: encrypted secret
        """
        if not isinstance(secret, dict):
            raise TypeError("secret must be a dict")
        
        with sfh.FileHandler(self.cryptkeys_filepath) as crypt_keys_file_hdl:
            try:
                cryptkeys: Dict = crypt_keys_file_hdl.read_file()
                m_key = cryptkeys.get('DJSM_MASTER_KEY', None)
                pub_key = cryptkeys.get('DJSM_PUBLIC_KEY', None)
            except Exception:
                m_key, pub_key = None, None
            priv_key = os.getenv('DJSM_PRIVATE_KEY')

            if all((m_key, pub_key, priv_key)) is False:
                key_signature = dcrypt.CryptKey.make_signature(hash_algorithm="SHA-256")
                m_key, pub_key, priv_key, _ = key_signature.common()
                cryptkeys = {
                    'DJSM_MASTER_KEY': m_key,
                    'DJSM_PUBLIC_KEY': pub_key,
                }
                crypt_keys_file_hdl.write_to_file(cryptkeys)
                del cryptkeys
                self._remove_priv_key_from_env()
                env_file_hdl = _get_env_hdl()
                env_file_hdl.write_to_file(f'DJSM_PRIVATE_KEY = "{priv_key}"\n', write_mode='a+')
                self.reload_env()

            common_signature = dcrypt.CommonSignature(m_key, pub_key, priv_key, "SHA-256")
            cryptkey = dcrypt.CryptKey(signature=dcrypt.Signature.from_common(common_signature))
            crypt = dcrypt.JSONCrypt(key=cryptkey)
            del m_key, pub_key, priv_key
            return crypt.encrypt(secret)


    def _decrypt(self, encrypted_secret: Dict) -> Dict:
        """
        Decrypts an encrypted secret using the encryption key.

        :param encrypted_secret: encrypted secret to be decrypted
        :return: decrypted secret
        """
        if not isinstance(encrypted_secret, dict):
            raise TypeError("encrypted_secret must be a dict")
        
        with sfh.FileHandler(self.cryptkeys_filepath) as crypt_keys_file_hdl:
            try:
                cryptkeys: Dict = crypt_keys_file_hdl.read_file()
                m_key: str = cryptkeys.get('DJSM_MASTER_KEY', None)
                pub_key: str = cryptkeys.get('DJSM_PUBLIC_KEY', None)
            except Exception:
                m_key, pub_key = None, None

            priv_key: str = os.getenv('DJSM_PRIVATE_KEY')

            if all((m_key, pub_key, priv_key)) is False:
                raise CryptKeysNotFound("Crypt key(s) not found. Secrets will not be recoverable")

            common_signature = dcrypt.CommonSignature(m_key, pub_key, priv_key, "SHA-256")
            cryptkey = dcrypt.CryptKey(signature=dcrypt.Signature.from_common(common_signature))
            crypt = dcrypt.JSONCrypt(key=cryptkey)
            del m_key, pub_key, priv_key
            return crypt.decrypt(encrypted_secret)
    

    @staticmethod
    def _remove_priv_key_from_env() -> None:
        """Removes rsa private key from .env file and environmental variables if present"""
        with _get_env_hdl() as env_file_handler:
            env_file_dict = dotenv_values(find_dotenv(raise_error_if_not_found=True))
            env_file_dict.pop('DJSM_PRIVATE_KEY', None)

            # Clear existing variables in .env file and re-write
            env_file_handler.clear_file()
            for key, value in env_file_dict.items():
                line = f'{key} = "{value}"\n'
                env_file_handler.write_to_file(line, write_mode='a+')
            os.environ.pop("DJSM_PRIVATE_KEY", None)
        return None


    def change_cryptkeys(self) -> None:
        """
        Change the encryption keys
        """
        # Get existing secrets
        secrets = self._load_secrets(self.secrets_file_path, decrypt=True)
        # Delete fernet and rsa public keys in JSON file - Just delete the cryptkeys file
        sfh.FileHandler(self.cryptkeys_filepath).delete_file()
        # Delete rsa private key in .env file
        self._remove_priv_key_from_env()
        # Re-write secrets (overwrite) while encrypting them with new keys
        self._write_secrets(secrets, self.secrets_file_path, encrypt=True, overwrite=True)
        del secrets
        return None
        

    def get_or_create_secret_key(self, always_generate: bool = False) -> str:
        """
        Gets Django secret key from secrets file, if it exists.
        Create and saves a secret key in secrets file if it cannot be found.

        The secret key will be saved to the path specified in the constructor or as specified in the
        .env file

        :param always_generate: if True, a new secret key will be generated and saved
        even if one already exists in the secret file.
        :return: secret key
        """ 
        secret_key = self.get_secret(self.secret_key_name)
        if not secret_key or always_generate:
            self.stdout("DJSM: Generating Secret Key...\n")
            secret_key = generate_django_secret_key()
            self.update_secrets({self.secret_key_name: secret_key})
            self.stdout("DJSM: Secret Key Generated Successfully!\n")
        else:
            self.stdout("DJSM: Secret Key Found!\n")

        if not validate_secret_key(secret_key):
            self.stdout("DJSM: Invalid Secret Key. Replacing Key...\n")
            # Replace secret key if the secret key is not valid and return the new secret key
            return self.get_or_create_secret_key(always_generate=True)
        return secret_key


    def _write_secrets(
        self, 
        secrets: Dict[str, Any], 
        path_to_file: str, 
        overwrite: bool = False, 
        encrypt: bool = False
    ) -> None:
        """
        Writes the secrets in the given path.

        :param secrets: dictionary containing new secrets.
        :param path_to_file: path to file in which secrets will be written into.
        :param overwrite: whether to overwrite all existing secrets in the file.
        :param encrypt: whether to encrypt the secrets before writing in the file. 
        :return: None.
        """
        if not isinstance(secrets, dict):
            raise TypeError('Secret must be a dictionary')

        with sfh.FileHandler(path_to_file) as file_hdl:
            if not file_hdl.filetype == "json":
                raise TypeError("Secret file must be a json file")

            if secrets:
                if encrypt:
                    secrets = self._encrypt(secrets)
                if overwrite:
                    file_hdl.write_to_file(secrets)
                else:
                    file_hdl.update_json(secrets)
                del secrets
        return None

    
    def _load_secrets(self, path_to_file: str, decrypt: bool = False) -> Dict:
        """
        Loads the secrets from the given path.

        :param path_to_file: path to file containing secrets.
        :param decrypt: whether to decrypt the secrets in the file before returning.
        :return: a dictionary of secrets.
        """
        with sfh.FileHandler(path_to_file) as file_hdl:
            try:
                if not file_hdl.filetype == "json":
                    raise TypeError("Secret file must be a json file")

                enc_secrets = file_hdl.read_file()
                if not isinstance(enc_secrets, dict):
                    raise TypeError('Secrets must be a dictionary')
                if enc_secrets and decrypt:
                    return self._decrypt(enc_secrets)
                return enc_secrets
            
            except sfh.FileError:
                return {}


    def get_secret(self, key: str) -> Any | None:
        """
        Gets secret from json secret(s) file.

        :param key: key to get from the json file(s)
        :return: secret if found, None if not
        """ 
        enc_secrets = self._load_secrets(self.secrets_file_path)
        enc_secret = enc_secrets.get(key, None)
        if enc_secret:
            dec_secret = self._decrypt({key: enc_secret})
            return dec_secret.get(key, None)
        return enc_secret


    def update_secrets(self, new_secrets: Dict[str, Any]) -> None:
        """
        Updates secrets in secrets file.

        :param new_secrets: secret to write to the json file
        :return: None
        """
        if not isinstance(new_secrets, dict):
            raise TypeError('Secret must be a dictionary')
        self._write_secrets(new_secrets, self.secrets_file_path, encrypt=True)
        del new_secrets
        return None


    def clean_up(self) -> None:
        """
        Deletes all secrets in the secrets file. Clears all environment variables
        set by the package.

        This method is useful when you want to delete all secrets and start afresh.
        The cryptkeys are left untouched.
        """
        sfh.FileHandler(self.secrets_file_path).delete_file()
        for var in env_variables:
            os.environ.pop(var, None)
        return None


    @staticmethod
    def reload_env() -> None:
        """
        Reloads the environment variables from the .env file.
        """
        find_and_load_env_var()
        return None


    def clean_up_and_reload(self) -> None:
        """
        Deletes all secrets in the secret file. Clears all environment variables
        set by the package and reloads the environment variables from the .env file.

        This method is useful when you want to delete all secrets and start afresh.
        """
        self.clean_up()
        self.reload_env()
        return None



# Just to allow use of alias "DJSM"
class DJSM(DjangoJSONSecretManager):
    """
    #### DJSM class for managing Django secrets in a json file.

    EXAMPLE:

    In settings.py:
    ```
    from djsm import DJSM

    djsm = DJSM('./.hidden_folder/secretfile.json')
    # SECURITY WARNING: keep the secret key used in production secret!
        # generate secret key if it does not exist
    SECRET_KEY = djsm.get_or_create_secret_key()
    ```
    """
    pass




from django.core.management.base import BaseCommand

import djsm


class Command(BaseCommand):
    """
    Delete the secrets  and remove environment variables set by secrets manager.

    Example:
    ```bash
    python manage.py djsm_cleanup
    ```
    """
    help = '''
    Delete the secrets (project secret key inclusive) and removes environment variables set by secrets manager.
    '''

    def handle(self, *args, **options) -> None:
        manager = djsm.get_djsm(quiet=True)
        manager.clean_up()
        self.stdout.write(
            self.style.SUCCESS('DJSM: Clean up completed!')
        )
        return None



from django.core.management.base import BaseCommand
from django.conf import settings, Settings
import simple_file_handler as sfh
from typing import List
import re
from django.utils.module_loading import import_string

from ...manager import DJSM


def get_settings_filepath() -> str | None:
    """Return the path to the project's settings file."""
    try:
        settings.configure()
    except RuntimeError:
        pass

    settings_module = settings.SETTINGS_MODULE
    if settings_module is None:
        return None
    return settings.SETTINGS_MODULE.split('.')[0] + '/settings.py'


def check_if_djsm_is_imported_in_settings(settings_lines: List[str]) -> bool:
    """
    Return True if djsm is already imported in the project's settings file.

    :param settings_lines: list containing the lines of the project's settings file.
    """
    possible_imports = ('djsm', 'DJSM', 'DjangoJSONSecretsManager')
    for line in settings_lines:
        # Ignore comments and blank lines and docstrings.
        if line.startswith(("#", "'''", '"""')) or line.strip() == '':
            continue
        # Ignore lines that are not imports.
        if not line.startswith(('import ', 'from ')):
            continue
        # If any of the possible imports are in the line, return True.
        if any(map(lambda x: x in line, possible_imports)):
            return True
    return False


def get_djsm_obj_name_in_settings() -> str | None:
    """
    Return the name of the DJSM object in the project's settings file, If there is one.

    If there is no DJSM object in the project's settings file, return None.
    """
    settings_module: Settings = import_string(settings.SETTINGS_MODULE)
    for name, obj in settings_module.__dict__.items():
        if isinstance(obj, DJSM):
            return name
    return None


def get_secret_key_line_index(settings_lines: List[str]) -> int:
    """
    Return the index of the line in the project's settings file that contains the secret key.

    :param settings_lines: list containing the lines of the project's settings file.
    :return: The index of the line in the project's settings file that contains the secret key.
    Returns -1 if there is no such line.
    """
    pattern = re.compile(r'^SECRET_KEY\s*=\s*.*\s*$')
    # Iterate through the lines in reverse order, 
    # so the last definition of the `SECRET_KEY` setting is replaced
    for i in range(len(settings_lines) - 1, -1, -1):
        line = settings_lines[i].strip()
        if pattern.match(line):
            return i
    return -1


def get_index_of_line_after_top_imports(settings_lines: List[str]) -> int:
    """
    Return the index of the line after the imports in the project's settings file.

    :param settings_lines: list containing the lines of the project's settings file.
    :return: The index of the line after the imports in the project's settings file.
    Returns 0 if there are no lines or imports in the project's settings file.

    NOTE: This does not factor in imports that a wrapped in a try-except block
    """
    for index, line in enumerate(settings_lines):
        line = line.strip()
        # If the line is empty, a comment or a docstring line, skip it
        if not line or line.startswith(("#", "'''", '"""')):
            continue
        # If it is an import line, skip it
        if line.startswith(('import ', 'from ')):
            continue
        return index
    return 0


def manage_secret_key_with_djsm(manager_name: str, quiet: bool, command: BaseCommand) -> None:
    """
    Edits project settings to use djsm to manage secret key. To be used by a `BaseCommand`.

    :param manager_name: The name of the DJSM object in the project's settings file.
    :param quiet: If set to True, DJSM will not print any messages.
    :param command: The BaseCommand object that called this function.
    """
    settings_filepath = get_settings_filepath()
    if settings_filepath is None:
        raise Exception('settings.SETTINGS_MODULE is not set.')
    
    with sfh.FileHandler(settings_filepath, not_found_ok=False, allow_any=True) as hdl:
        content = hdl.file_content
        lines = content.split('\n')
        project_name = settings.SETTINGS_MODULE.split('.')[0]
        has_been_imported = check_if_djsm_is_imported_in_settings(lines)
        djsm_obj_name = get_djsm_obj_name_in_settings()
        secret_key_line_index = get_secret_key_line_index(lines)
        index_after_top_imports = get_index_of_line_after_top_imports(lines)
        secret_key_line = f'''SECRET_KEY = {manager_name}.get_or_create_secret_key()'''

        if has_been_imported is True and djsm_obj_name is not None:
            pattern = re.compile(
                f'''^SECRET_KEY\\s*=\\s*{djsm_obj_name}\\.[a-zA-Z_]*\\(([\\'"]+[\\w]*[\\'"]+)?\\)\\s*$'''
            )
            # If the secret key line is already managed by DJSM, do nothing.
            if pattern.match(lines[secret_key_line_index]):
                command.stdout.write(
                    command.style.SUCCESS(f'DJSM: Looks like DJSM already manages the secret key for {project_name}!')
                )
                return
            lines[secret_key_line_index] = f'''SECRET_KEY = {djsm_obj_name}.get_or_create_secret_key()'''

        elif has_been_imported is True and djsm_obj_name is None:
            setup_line = f"""{manager_name} = djsm.get_djsm(quiet={quiet})"""
            lines.insert(index_after_top_imports, setup_line)
            lines[secret_key_line_index] = secret_key_line

        elif has_been_imported is False and djsm_obj_name is None:
            # Import package at the top of the file
            lines.insert(0, "import djsm")
            setup_line = f"""\n{manager_name} = djsm.get_djsm(quiet={quiet})"""
            # Setup manager after import
            lines.insert(index_after_top_imports, setup_line)
            lines[secret_key_line_index] = secret_key_line

        new_content = '\n'.join(lines)
        # Overwrite the project's settings file with the new content.
        hdl.write_to_file(new_content, write_mode='w')

    command.stdout.write(
        command.style.SUCCESS(f'DJSM: DJSM now manages the secret key for {project_name}!')
    )
    return None



class Command(BaseCommand):
    """
    Use DJSM to manage the secret key in the project's settings file.

    This command will add the DJSM object to the project's settings file if it is not already there.

    Example usage:
    ```bash
    python manage.py use_djsm --as djsm_manager --quiet
    ```

    The following code will be added to the project's settings file:
    ```python
    ...
    import djsm

    djsm_manager = djsm.get_djsm(quiet=True)
    ...
    SECRET_KEY = djsm_manager.get_or_create_secret_key()
    ...
    """
    help = 'Use DJSM to manage the secret key in the project\'s settings file.'

    def add_arguments(self, parser) -> None:
        parser.add_argument(
            '--as',
            type=str,
            default='djsm_manager',
            help='The name to be used for the DJSM object in the project\'s settings file.',
        )

        parser.add_argument(
            '--quiet',
            action='store_true',
            help='If set to True, DJSM will not print any messages.',
        )
        

    def handle(self, *args, **options) -> None:
        manager_name = options['as']
        quiet = options['quiet']
        manage_secret_key_with_djsm(manager_name, quiet=quiet, command=self)
        return None



import json
from django.core.management.base import BaseCommand, CommandParser

import djsm


def parse_value(value: str) -> str:
    try:
        return json.loads(value)
    except json.JSONDecodeError:
        return value
    

class Command(BaseCommand):
    """
    Update secrets file with new secrets. The secrets file updated is the one defined in the .env file.

    Example:
    ```bash
    python manage.py update_secrets API_KEY 1234567890
    ```
    """
    help = '''
    Update secrets file with new secrets. The secrets file updated is the one defined in the .env file.
    '''

    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            'name',
            type=str,
            help='The name of the secret to be updated.',
        )

        parser.add_argument(
            'value',
            type=parse_value,
            help='The value of the secret to be updated.',
        )


    def handle(self, *args, **options) -> None:
        name = options['name']
        value = options['value']

        manager = djsm.get_djsm(quiet=True)
        manager.update_secrets({name: value})
        self.stdout.write(
            self.style.SUCCESS('DJSM: Secrets updated successfully!')
        )
        return None



"""
#### djsm - Django JSON Secrets Manager

Create, store, retrieve, update and manage secrets in your Django project.

@Author: Daniel T. Afolayan (ti-oluwa.github.io)
"""

__version__ = "0.1.51"
__author__ = "Daniel T. Afolayan (ti-oluwa)"
__license__ = "MIT"

import os
import sys

from .manager import DJSM
from .misc import check_setup

_djsm_setup_checked = False


def get_djsm(quiet: bool = False):
    """
    Returns an instance of the DJSM class instantiated based on the environment variables set.

    :param quiet: If True, do not write anything to stdout (to avoid cluttering the console)
    """
    global _djsm_setup_checked
    if not _djsm_setup_checked:
        if not check_setup(quiet):
            sys.exit()
        _djsm_setup_checked = True
    return DJSM(os.getenv('DJSM_SECRETS_FILE_PATH'), quiet=quiet)





from django.core.management.base import BaseCommand

import djsm


class Command(BaseCommand):
    """
    Changes the encryption and decryption keys used by secrets manager.

    Example:
    ```bash
    python manage.py change_cryptkeys
    ```
    """
    help = '''
    Changes the encryption and decryption keys used by secrets manager.
    '''

    def handle(self, *args, **options) -> None:
        manager = djsm.get_djsm(quiet=True)
        manager.change_cryptkeys()
        self.stdout.write(
            self.style.SUCCESS('DJSM: Encryption keys change was successfully!')
        )
        return None



import string
from dotenv import load_dotenv, find_dotenv, dotenv_values
import sys
import random

from .exceptions import EnvLoadError


SECRET_KEY_ALLOWED_CHARS = string.ascii_letters + string.digits + string.punctuation

env_variables = (
    "DJSM_SECRETS_FILE_PATH",
    "DJSM_SECRET_KEY_NAME",
)


def find_and_load_env_var() -> None:
    """Load environment variables from .env file"""
    try:
        load_dotenv(find_dotenv('.env', raise_error_if_not_found=True), override=True)
    except Exception:
        raise EnvLoadError("Could not load environmental variables because '.env' file was not found. Create one!")
    return None


def _print(msg: str, quiet: bool = False) -> None:
    if not quiet:
        sys.stdout.write(msg)


def check_setup(quiet: bool = False) -> bool:
    """
    Check that an .env file is present and has been properly setup
    
    :param quiet: If True, do not write anything to stdout (to avoid cluttering the console)
    :return: True if setup is OK, False if not
    """
    setup_ok = True
    _print("DJSM: Ensuring setup is OK.\n", quiet=quiet)
    try:
        _print("DJSM: Searching for .env file...\n", quiet=quiet)
        dotenv_path = find_dotenv(raise_error_if_not_found=True)
    except IOError:
        setup_ok = False
        _print("DJSM: Could not load or find .env file!\n")

    if setup_ok:
        _print("DJSM: Checking that .env file has been properly setup...\n", quiet=quiet)
        env_file_dict = dotenv_values(dotenv_path)
        if not env_file_dict.get('DJSM_SECRETS_FILE_PATH', None):
            _print('DJSM: DJSM_SECRETS_FILE_PATH not set in .env file\n')
            setup_ok = False
        
    if setup_ok:
        load_dotenv(dotenv_path, override=True)
        _print('DJSM: Setup OK!\n', quiet=quiet)
    else:
        _print('DJSM: Visit https://github.com/ti-oluwa/djsm/#usage for help on how to setup DJSM\n')
    return setup_ok
 

def validate_secret_key(secret_key: str) -> bool:
    """Check that Django secret key is valid."""
    if not isinstance(secret_key, str):
        return False
    if len(secret_key) < 32:
        return False
    for char in secret_key:
        if char not in SECRET_KEY_ALLOWED_CHARS:
            return False
    return True


def generate_django_secret_key(length: int = 50) -> str:
    """
    Return a randomly generated key of not less than 32 characters
    
    :param length: length of secret key to be generated.
    """
    if length < 32:
        raise ValueError('Secret key length cannot be less than 32 characters')
    return ''.join(random.choice(SECRET_KEY_ALLOWED_CHARS) for _ in range(length))
